// © 2023 John Breaux
//! Traverses an AST, assembling instructions.
//!
//! [Assembler] carries *some* state

use crate::parser::preamble::*;
use error::AssemblyError;
use std::collections::HashMap;
use std::path::Path;

pub mod error;

#[derive(Clone, Debug, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub enum IdentType {
    Word,
    Jump,
}

/// Takes in an AST's [Root], and outputs a sequence of bytes
#[derive(Clone, Debug, Default, PartialEq, Eq)]
pub struct Assembler {
    out: Vec<u16>,
    /// A map from Labels' [Identifier]s to their location in the binary
    labels: HashMap<Identifier, usize>,
    /// A list of all referenced [Identifier]s in the binary, and their locations
    identifiers: Vec<(usize, Identifier, IdentType)>,
}

impl Assembler {
    pub fn assemble(r: &Root) -> Result<Vec<u16>, AssemblyError> {
        let mut out = Self::default();
        out.visit_root(r)?;
        Ok(out.out)
    }
    pub fn load(&mut self, r: &Root) -> Result<(), AssemblyError> { self.visit_root(r) }
    pub fn out(self) -> Vec<u16> { self.out }
    
    fn last_mut(&mut self) -> Result<&mut u16, AssemblyError> { self.out.last_mut().ok_or(AssemblyError::EmptyBuffer) }
    fn push_default(&mut self) -> usize {
        self.out.push(Default::default());
        self.out.len() - 1
    }
}

impl Assembler {
    /// Visits the [Root] node of a parse tree
    fn visit_root(&mut self, r: &Root) -> Result<(), AssemblyError> {
        // Visit the entire tree
        for (num, line) in r.lines() {
            self.visit_line(line).map_err(|e| e.ctx(r.file().unwrap_or(Path::new("stdin")), *num))?;
        }
        // Link identifiers
        for (idx, id, id_type) in self.identifiers.iter() {
            let Some(&num) = self.labels.get(id) else { return Err(AssemblyError::UnresolvedIdentifier(id.clone())) };
            let offset = (num as isize - *idx as isize) * 2;
            *self.out.get_mut(*idx).expect("idx should be a valid index into out") |= match id_type {
                IdentType::Word => offset as u16,
                IdentType::Jump => JumpTarget::squish(offset)?,
            };
        }
        Ok(())
    }

    /// visit a [Line]
    fn visit_line(&mut self, line: &Line) -> Result<(), AssemblyError> {
        match line {
            Line::Insn(insn) => self.visit_instruction(insn),
            Line::Label(label) => self.visit_label(label),
            Line::Directive(d) => self.visit_directive(d),
            _ => Ok(()),
        }
    }

    /// Visits a [Directive]
    fn visit_directive(&mut self, node: &Directive) -> Result<(), AssemblyError> {
        match node {
            Directive::Org(_) => todo!("{node}"),
            Directive::Define(..) => (),
            Directive::Include(r) => self.visit_root(r)?,
            Directive::Byte(word) | Directive::Word(word) => self.out.push((*word).into()),
            Directive::Bytes(words) | Directive::Words(words) => {
                for word in words {
                    self.out.push((*word).into());
                }
            }
            Directive::String(s) => self.visit_string(s)?,
            Directive::Strings(strs) => {
                for s in strs {
                    self.visit_string(s)?;
                }
            }
        };
        Ok(())
    }

    /// Visits a [Label]
    fn visit_label(&mut self, node: &Label) -> Result<(), AssemblyError> {
        // Register the label
        match self.labels.insert(node.0.to_owned(), self.out.len()) {
            Some(_) => Err(AssemblyError::RedefinedLabel(node.0.to_owned())),
            _ => Ok(()),
        }
    }

    /// Visits an [Instruction]
    fn visit_instruction(&mut self, insn: &Instruction) -> Result<(), AssemblyError> {
        self.push_default();
        self.visit_opcode(insn.opcode())?;
        self.visit_encoding(insn.encoding())?;
        Ok(())
    }

    /// Visits an [Opcode]
    fn visit_opcode(&mut self, node: &Opcode) -> Result<(), AssemblyError> {
        *self.last_mut()? |= *node as u16;
        Ok(())
    }

    /// Visits an [Encoding]
    fn visit_encoding(&mut self, node: &Encoding) -> Result<(), AssemblyError> {
        *self.last_mut()? |= node.word();
        match node {
            Encoding::Single { dst, .. } => {
                self.visit_primary_operand(dst)?;
            }
            Encoding::Jump { target } => {
                self.visit_jump_target(target)?;
            }
            Encoding::Double { src, dst, .. } => {
                self.visit_primary_operand(src)?;
                self.visit_secondary_operand(dst)?;
            }
        }
        Ok(())
    }

    /// Visits a [JumpTarget]
    fn visit_jump_target(&mut self, node: &JumpTarget) -> Result<(), AssemblyError> {
        match node {
            JumpTarget::Number(num) => self.visit_number(num),
            JumpTarget::Identifier(id) => {
                self.visit_identifier(id, self.out.len() - 1, IdentType::Jump)?;
                Ok(())
            }
        }
    }

    /// Visits a [SecondaryOperand]
    fn visit_secondary_operand(&mut self, node: &SecondaryOperand) -> Result<(), AssemblyError> {
        use SecondaryOperand as O;
        if let O::Indexed(_, num) | O::Absolute(num) = node {
            self.push_default();
            self.visit_number(num)?;
        }
        Ok(())
    }

    /// Visits a [PrimaryOperand]
    fn visit_primary_operand(&mut self, node: &PrimaryOperand) -> Result<(), AssemblyError> {
        use PrimaryOperand as O;
        match node {
            O::Indexed(_, num) | O::Absolute(num) | O::Immediate(num) => {
                self.push_default();
                self.visit_number(num)?;
            }
            O::Relative(id) => {
                let addr = self.push_default();
                self.visit_identifier(id, addr, IdentType::Word)?;
            }
            _ => (),
        }
        Ok(())
    }

    /// Visits a number and writes it into the last index
    fn visit_number(&mut self, node: &Number) -> Result<(), AssemblyError> {
        *self.last_mut()? |= u16::from(*node);
        Ok(())
    }

    /// Visits a number and appends it to the output buffer
    fn visit_string(&mut self, node: &str) -> Result<(), AssemblyError> {
        for (idx, byte) in node.bytes().chain([0u8].into_iter()).enumerate() {
            if idx % 2 == 0 {
                self.push_default();
            }
            *self.last_mut()? |= (byte as u16) << (8 * (idx % 2));
        }
        Ok(())
    }

    /// Visits an [Identifier], and registers it to the identifier list
    fn visit_identifier(&mut self, node: &Identifier, addr: usize, ty: IdentType) -> Result<(), AssemblyError> {
        self.identifiers.push((addr, node.clone(), ty));
        Ok(())
    }
}
