// © 2023 John Breaux
//! Parses [`Tokens`](crate::Token) into an [abstract syntax tree](Root)

use crate::{TokenStream, Type};
use error::ParseError;
use preamble::*;
use std::{
    fmt::{Debug, Display},
    path::Path,
};

pub mod preamble {
    //! All the different AST node types
    use super::*;
    // Traits
    pub use parsable::Parsable;
    // Nodes
    pub use comment::Comment;
    pub use directive::Directive;
    pub use identifier::Identifier;
    pub use instruction::{
        encoding::{
            encoding_parser::EncodingParser, jump_target::JumpTarget, number::Number, primary_operand::PrimaryOperand,
            register::Register, secondary_operand::SecondaryOperand, width::Width, Encoding,
        },
        opcode::Opcode,
        Instruction,
    };
    pub use label::Label;
    pub use line::Line;
    pub use root::Root;
    // Error
    pub use error::ParseError;
}

pub mod parsable;

pub mod comment;
pub mod directive;
pub mod error;
pub mod identifier;
pub mod instruction;
pub mod label;
pub mod line;
pub mod root;

pub struct Parser {
    radix: u32,
}

impl Parser {
    pub fn parse_with<'t>(self, stream: &'t mut impl TokenStream<'t>) -> Result<Root, ParseError> {
        Root::parse(&self, &mut stream.ignore(Type::Space))
    }
    pub fn parse<T>(self, input: &T) -> Result<Root, ParseError>
    where T: AsRef<str> + ?Sized {
        Root::parse(&self, &mut super::Tokenizer::new(input).preprocessed().ignore(Type::Space))
    }
    pub fn parse_file<P>(self, path: &P) -> Result<Root, ParseError>
    where P: AsRef<Path> + ?Sized {
        self.parse(&std::fs::read_to_string(path.as_ref())?).map(|r| r.set_file(path.as_ref().into()))
    }
    pub fn parse_one<T>(self, input: &T) -> Result<Line, ParseError>
    where T: AsRef<str> + ?Sized {
        Line::parse(&self, &mut super::Tokenizer::new(input).preprocessed().ignore(Type::Space))
    }

    /// Sets the default radix for [Token](crate::lexer::token::Token) -> [Number]
    /// conversion
    pub fn radix(mut self, radix: u32) { self.radix = radix; }
}

impl Default for Parser {
    fn default() -> Self { Self { radix: 16 } }
}

impl Debug for Parser {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("Parser").field("radix", &self.radix).finish_non_exhaustive()
    }
}
